/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: linux-2.4.6-iscsi.patch,v 1.2 2001/07/18 22:32:57 arjanv Exp $ 
 *
 * nuscsitcp.c
 *
 *      Linux iSCSI low-level scsi driver module
 * 
 * May 2000, MTM 
 *
 */

/* there's got to be a better way to wait for child processes created by kernel_thread */
static int errno = 0;
#define __KERNEL_SYSCALLS__

#define LINUXVERSION(V, P, S)    (((V) * 65536) + ((P) * 256) + (S))

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <linux/stddef.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/unistd.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/socket.h>

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
#include <asm/semaphore.h>
#else
#include <asm/spinlock.h>
#endif
#include <asm/uaccess.h>
#include <scsi/sg.h>
#include <endian.h>

#include "sd.h"
#include "scsi.h"
#include "hosts.h"

/* #define DEBUG_INIT   1 */
/* #define DEBUG_FLOW   1 */
#define DEBUG_ERROR 1
/* #define DEBUG_TRACE  1 */

#define USE_DYNAMIC_DEVICE

#include "scsiTcp.h"
#include "nuscsitcp.h"



MODULE_AUTHOR("Cisco Systems, Inc.");
MODULE_DESCRIPTION("iSCSI Driver");

static NuScsiTcpCmdEntry *AllocMyCmd( NuScsiTcpHostAdapter *ha,
                                      ConnInfoEntry *cie );
static void FreeMyCmd( NuScsiTcpCmdEntry *ce );
int NuScsiTcpTimer( void );
int NuScsiTcpConnection( NuScsiTcpIoctlData *ioctld );
int NuScsiTcpXmitCmd(ConnInfoEntry *cie, Scsi_Cmnd *sc, NuScsiTcpCmdEntry *ce);
void NuScsiTcpRecvCmd( struct ScsiTcpScsiRspHdr *stsrh, unsigned char *xbuf );
void NuScsiTcpQueueRttData( struct ScsiTcpRttHdr *strh );
void NuScsiTcpXmitRttData( ConnInfoEntry *cie, NuScsiTcpCmdEntry *ce );
void NuScsiTcpXmitImmData( ConnInfoEntry *cie, NuScsiTcpCmdEntry *ce );
void NuScsiTcpFillScsiCmd(ConnInfoEntry *cie, struct ScsiTcpDataRspHdr *stdrh);
int NuScsiTcpXmitMgmtMsg( NuScsiTcpCmdEntry *mycmd, Scsi_Cmnd *sc );
void NuScsiTcpMgmtResp( struct ScsiTcpScsiTaskMgtRspHdr *ststmrh );
void NuScsiTcpAsyncEvent( struct ScsiTcpAsyncEvtHdr *staeh );
void NuScsiTcpXmitPing( ConnInfoEntry *cie );
static int ControlOpen( struct inode *inode, struct file *file );
static int ControlClose( struct inode *inode, struct file *file );
static int ControlIoctl( struct inode *inode,
                         struct file *file,
                         unsigned int cmd,
                         unsigned long arg );
void NuScsiTcpFillTrace( unsigned char flag,
                         unsigned char lun,
                         unsigned char cmd,
                         unsigned char target,
                         unsigned int itt,
                         unsigned int data );
void NuScsiTcpXmitQdCmds( ConnInfoEntry *cie, Scsi_Cmnd *scq );
int NuScsiTcpConnTask( void *vtask );
static int MySendMsg( ConnInfoEntry *cie, struct msghdr *msg, int len );
static int MyRecvMsg( ConnInfoEntry *cie, struct msghdr *msg, int len );
void NuScsiTcpXmitNop( ConnInfoEntry *cie, struct ScsiTcpNopInHdr *stnih );

static int ControlMajor;
static const char *ControlName = "NuScsiTcpCtl";
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
static struct file_operations ControlFops = {
    owner: THIS_MODULE,
    ioctl: ControlIoctl,    /* ioctl */
    open: ControlOpen,      /* open */
    release: ControlClose,  /* release */
};
#else
static struct file_operations ControlFops = {
    NULL,                   /* lseek */
    NULL,                   /* read */
    NULL,                   /* write */
    NULL,                   /* readdir */
    NULL,                   /* poll */
    ControlIoctl,       /* ioctl */
    NULL,                   /* mmap */
    ControlOpen,        /* open */
    NULL,                   /* flush */
    ControlClose,       /* release */
};
#endif

static NuScsiTcpHostAdapter *HostA;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
static wait_queue_head_t WaitQ;
#else
static struct wait_queue *WaitQ = NULL;
#endif
static int TimerActive = 0;
static int NumConnections = 0;
static int NumTargets = 0;
static int ActiveCmds = 0;
spinlock_t NuScsiTcpSpinLock = SPIN_LOCK_UNLOCKED;
static ConnInfoEntry *ConnInfoHead = NULL;
static ConnInfoEntry *ConnInfoTail = NULL;
static ConnInfoEntry *ConnInfo[NUSCSITCP_MAX_TARGETS];
static unsigned char TargetIds[NUSCSITCP_MAX_TARGETS];
static char TargetNames[NUSCSITCP_MAX_TARGETS][16];
static int TargetLuns[NUSCSITCP_MAX_TARGETS];

#ifdef DEBUG_TRACE
static NuScsiTcpTraceEntry TraceTable[NUSCSITCP_TRACE_COUNT];
static int TraceIndex=0;
#define NU_TRACE( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT ) \
    NuScsiTcpFillTrace( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT );
#else
#define NU_TRACE( P_FLG, P_LUN, P_CMD, P_TGT, P_ITT, P_DAT )
#endif

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,27) )
/* note change modeled per linux2.4 drivers/scsi/ips.c */
struct proc_dir_entry proc_dir_nuscsitcp = {
#ifdef PROC_SCSI_ISCSI
    PROC_SCSI_ISCSI,
#else
    PROC_SCSI_NOT_PRESENT,
#endif
    9,
    "NuScsiTcp",
    S_IFDIR|S_IRUGO|S_IXUGO,
    2
};
#endif

int
nuscsitcp_detect( Scsi_Host_Template *sht )
{
    struct Scsi_Host *sh;
    NuScsiTcpHostAdapter *ha;
    NuScsiTcpCmdEntry *ce;
    int i;

    DEBUG_INIT0("iSCSI: Detect\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    sht->proc_dir = &proc_dir_nuscsitcp;
#else
    sht->proc_name = "NuScsiTcp";
#endif

    sh = scsi_register( sht, sizeof(NuScsiTcpHostAdapter) );
    if ( ! sh ) {
        printk("iSCSI: Unable to register controller\n");
        return 0;
    }

#ifdef USE_DYNAMIC_DEVICE
    sh->max_id = NUSCSITCP_MAX_TARGETS;
#else
    sh->max_id = NumTargets;
#endif
    sh->max_lun = NUSCSITCP_MAX_LUN;
    sh->max_channel = 0;

    ha = (NuScsiTcpHostAdapter *)sh->hostdata;
    memset( ha, 0, sizeof(NuScsiTcpHostAdapter) );

    i = (NUSCSITCP_CANQUEUE * sizeof(NuScsiTcpCmdEntry));
    ce = kmalloc( i, GFP_KERNEL);
    if ( ! ce ) {
        printk("iSCSI: Detect could not allocate memory\n");
        return 0;
    }
    memset( ce, 0, i);
    ha->cmd_mem = ce;
    ha->free_cmd_head = ce;
    for( i=1; i<NUSCSITCP_CANQUEUE; i++ ) {
        ce->next = (ce + 1);
        ce->next->prev = ce;
        ce = ce->next;
        ce->next = NULL;
    }

    ha->free_cmd_tail = ce;
    ha->cmd_max = ce;

    ha->active = 1;

    HostA = ha;

    return 1;   /* one host detected */
}

int
nuscsitcp_release( struct Scsi_Host *sh )
{
    NuScsiTcpHostAdapter *ha;
    ConnInfoEntry *cie;

    DEBUG_INIT0("iSCSI: Release\n");

    cie = ConnInfoHead;
    while( cie ) {
        ConnInfoHead = cie->Next;
        kfree( cie );
        cie = ConnInfoHead;
    }

    ha = (NuScsiTcpHostAdapter *)sh->hostdata;
    if ( ! ha ) {
        return FALSE;
    }
    kfree( ha->cmd_mem );
    memset( ha, 0, sizeof(NuScsiTcpHostAdapter) );
    
    scsi_unregister( sh );
    HostA = NULL;

    return FAILED;
}

#ifdef ISCSI_USE_STRATEGY
int
nuscsitcp_strategy( struct Scsi_Host *host )
{
    Scsi_Device *sd;
    Scsi_Cmnd *sc;
    ConnInfoEntry *cie;
    NuScsiTcpCmdEntry *ce;

    for( sd=host->host_queue; sd; sd=sd->next ) {
        printk("iSCSI: Strategy Device 0x%x, Id %d, Lun %d, Chan %d\n",(int)sd,sd->id,sd->lun,sd->channel);
        for( sc=sd->device_queue; sc; sc=sc->next ) {
            printk("iSCSI: Strategy Cmd 0x%x, state 0x%x, tgt 0x%x\n",(int)sc,sc->state,sc->target);
        }
    }

    sc = HostA->q_head;
    while( sc ) {
        printk("iSCSI: qHead 0x%x\n",(int)sc);
        sc = (Scsi_Cmnd *)sc->host_scribble;
    }

    cie = ConnInfoHead;
    while( cie ) {
        sc = cie->ConnScHead;
        while( sc ) {
            printk("iSCSI: ConnScHead 0x%x\n",(int)sc);
            sc = (Scsi_Cmnd *)sc->host_scribble;
        }
        ce = cie->CmdHead;
        while( ce ) {
            printk("iSCSI: CmdHead 0x%x\n",(int)ce->sc);
            ce = ce->next;
        }
        cie = cie->Next;
    }

    return 0;
}
#endif

int
nuscsitcp_eh_abort( Scsi_Cmnd *sc )
{
    NuScsiTcpCmdEntry *mycmd;
    int rc;

    if ( ! sc ) {
        return FAILED;
    }

    sc->result = DID_ABORT << 16;

    mycmd = (NuScsiTcpCmdEntry *)sc->host_scribble;
    if ( (mycmd >= HostA->cmd_mem) &&
         (mycmd <= HostA->cmd_max) &&
         (mycmd->sc == sc) ) {
        mycmd->cmdaborted = 1;
        printk("iSCSI: abort command 0x%x, Tgt %d, Lun %d, Cmd 0x%x\n",
            (unsigned int)mycmd,sc->target,sc->lun,sc->cmnd[0]);
        DEBUG_INIT5(
            "iSCSI: EhAbort Cmd 0x%x, Len 0x%x, Rxd 0x%x, Rxr %d, Itt 0x%x\n",
            sc->cmnd[0],sc->request_bufflen,mycmd->rxdata,mycmd->rxrsp,
            (unsigned int)mycmd);
        DEBUG_INIT1("           Timeout %lu jiffies\n",(jiffies-mycmd->jiffies));
        if ( sc->eh_state == FAILED ) {
            printk("iSCSI: send abort, 0x%x, to remote\n",(unsigned int)mycmd);
            rc = NuScsiTcpXmitMgmtMsg( mycmd, sc );
            if ( rc ) {
                DEBUG_INIT0("iSCSI: EhAbort Xmit failed\n");
                return FAILED;
            }
            NU_TRACE( NUSCSITCP_TRACE_TxAbort, sc->lun, sc->cmnd[0], sc->target,
                (unsigned int)mycmd, mycmd->rxdata );
            current->state = TASK_INTERRUPTIBLE;
            rc = schedule_timeout( HZ );
            if ( mycmd->abortdone ) {
                printk("iSCSI: abort successful 0x%x\n",(unsigned int)mycmd);
                NU_TRACE( NUSCSITCP_TRACE_CmdAborted, sc->lun, sc->cmnd[0],
                    sc->target, (unsigned int)mycmd, mycmd->rxrsp );
                FreeMyCmd( mycmd );
                sc->host_scribble = NULL;
                return SUCCESS;
            }
        }
        FreeMyCmd( mycmd );
        sc->host_scribble = NULL;
    }
    return FAILED;
}

int
nuscsitcp_eh_host_reset( Scsi_Cmnd *sc )
{
    ConnInfoEntry *cie;

    DEBUG_INIT1("iSCSI: EhHostReset 0x%x\n",(int)sc);
    printk("iSCSI: host reset 0x%x\n",(unsigned int)sc);
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhHostReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }

    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout( (30*HZ) );
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhHostReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }
    
    DEBUG_INIT1("iSCSI: EhHostReset 0x%x FAILED\n",(int)sc);
    return FAILED;
}

int
nuscsitcp_eh_bus_reset( Scsi_Cmnd *sc )
{
    ConnInfoEntry *cie;

    printk("iSCSI: bus reset 0x%x\n",(unsigned int)sc);
    DEBUG_INIT1("iSCSI: EhBusReset 0x%x\n",(int)sc);
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhBusReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }

    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout( (4*HZ) );
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhBusReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }
    
    DEBUG_INIT1("iSCSI: EhBusReset 0x%x FAILED\n",(int)sc);
    return FAILED;
}

int
nuscsitcp_eh_device_reset( Scsi_Cmnd *sc )
{
    ConnInfoEntry *cie;

    printk("iSCSI: device reset 0x%x\n",(unsigned int)sc);
    DEBUG_INIT1("iSCSI: EhDeviceReset 0x%x\n",(int)sc);
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhDeviceReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }

    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout( (30*HZ) );
    cie = ConnInfo[sc->target];
    if ( cie && (cie->ConnFlag & CONN_ALIVE) ) {
        DEBUG_INIT1("iSCSI: EhDeviceReset 0x%x SUCCESS\n",(int)sc);
        return SUCCESS;
    }

    DEBUG_INIT1("iSCSI: EhDeviceReset 0x%x FAILED\n",(int)sc);
    return FAILED;
}

int
nuscsitcp_queue( Scsi_Cmnd *sc, void (*done)(Scsi_Cmnd *) )
{
    NuScsiTcpHostAdapter *ha;

    DEBUG_FLOW4("iSCSI: Q Target %d, Lun %d, Chan %d, Cmd 0x%x\n",sc->target,
        sc->lun,sc->channel,sc->cmnd[0]);
    ha = (NuScsiTcpHostAdapter *)sc->host->hostdata;
    if ( (!ha) || (!ha->active) ) {
        return 1;
    }

    if ( ! TimerActive ) {
        /* do not have any handler yet, fail the request */
        sc->result = DID_NO_CONNECT << 16;
        done( sc );
        return 0;
    }

    sc->scsi_done = done;
    sc->result = 0;
    sc->host_scribble = NULL;
    NU_TRACE( NUSCSITCP_TRACE_Qd, sc->lun, sc->cmnd[0], sc->target, 0,
        (unsigned int)sc );

    if ( ha->q_head ) {
        ha->q_tail->host_scribble = (char *)sc;
    } else {
        ha->q_head = sc;
    }
    ha->q_tail = sc;

    wake_up_interruptible( &WaitQ );

    return 0; /* queued, return OK */
}

int
nuscsitcp_biosparam( Disk *disk, kdev_t dev, int geom[] )
{
    DEBUG_INIT1("iSCSI: BiosParam, capacity %d\n",disk->capacity);
    geom[0] = 64;                                   /* heads */
    geom[1] = 32;                                   /* sectors */
    geom[2] = disk->capacity / (64*32);     /* cylinders */
    return 1;
}

const char *
nuscsitcp_info( struct Scsi_Host *sh )
{
    NuScsiTcpHostAdapter *ha;
    static char buffer[256];
    char *bp;

    DEBUG_INIT0("iSCSI: Info\n");
    ha = (NuScsiTcpHostAdapter *)sh->hostdata;
    if ( ! ha ) {
        return NULL;
    }

    bp = &buffer[0];
    memset( bp, 0, sizeof(buffer) );
#if defined(INTERNAL_DRIVER_VERSION) && (INTERNAL_DRIVER_VERSION > 0)        
    sprintf(bp, "iSCSI (%d.%d.%d.%d)", 
                (SCSI_TCP_ISCSI_VERSION+1), SCSI_TCP_CISCO_VERSION, SCSI_TCP_DRIVER_VERSION, INTERNAL_DRIVER_VERSION);
#else
    sprintf(bp, "iSCSI (%d.%d.%d)", 
                (SCSI_TCP_ISCSI_VERSION+1), SCSI_TCP_CISCO_VERSION, SCSI_TCP_DRIVER_VERSION );
#endif

    return bp;
}
int nuscsitcp_proc_info( char *buffer,
                     char **start,
                     off_t offset,
                     int length,
                     int hostno,
                     int func )
{
    int t, l, lfound;
    char *bp;
    ConnInfoEntry *cie;
    union {
        unsigned int l;
        unsigned char c[4];
    } ltoc;

    DEBUG_INIT4(
        "iSCSI: ProcInfo Offset 0x%lx, Len 0x%x, HostNo %d, Func %d\n",
        offset,length,hostno,func);

    bp = buffer;
    bp += sprintf(bp,"# iSCSI:\n");
    bp += sprintf(bp,"# Local: Hst Chn Tgt Lun     Remote:  IpAddr    Tgt Name\n");
    for( t=0; t<NumTargets; t++ ) {
            cie = ConnInfo[t];
            if ( cie ) {
                /* if we've already found LUNs, show them all */
                lfound = 0;
                for( l=0; l<NUSCSITCP_MAX_LUN; l++ ) {
                    if ( TargetLuns[t] & (1<<l) ) {
                        ltoc.l = cie->RemoteIpAddr;
                        bp += sprintf(bp, "         %03x 000 %03x %03x           %u.%u.%u.%u %03x %s\n", 
                                      hostno,t,l, 
                                      ltoc.c[0],ltoc.c[1],ltoc.c[2],ltoc.c[3], TargetIds[t],TargetNames[t]);
                        lfound += 1;
                    }
                }
                /* if we haven't found any LUNs, use ??? to tell nuactlun to probe for them */
                if ( ! lfound ) {
                    ltoc.l = cie->RemoteIpAddr;
                    bp += sprintf(bp, "         %03x 000 %03x ???            %u.%u.%u.%u %03x %s\n", 
                                  hostno,t,
                                  ltoc.c[0],ltoc.c[1],ltoc.c[2],ltoc.c[3], TargetIds[t],TargetNames[t]);
                }
            }
    }
    
    *start = buffer + offset;

    if ( (bp-buffer) < offset ) {
        return 0;
    }

    if ( (bp-buffer-offset) < length ) {
        return (bp-buffer-offset);
    }

    return length;
}

/*
 * We cannot include scsi_module.c because the daemon has not got a connection
 * up yet.
 */
static int
ControlOpen( struct inode *inode, struct file *file )
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int
ControlClose( struct inode *inode, struct file *file )
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int
ControlIoctl( struct inode *inode,
              struct file *file,
              unsigned int cmd,
              unsigned long arg)
{
    int rc;
    NuScsiTcpIoctlData ioctld;

    if ( ! current->files ) {
        printk("iSCSI: driver does not match kernel\n");
        return -EFAULT;
    }

    if ( cmd == NUSCSITCP_GO ) {
        if ( TimerActive ) {
            return -EBUSY;
        }
        rc = NuScsiTcpTimer();
        return rc;
    }

    if ( cmd == NUSCSITCP_CONN ) {
        if ( copy_from_user(&ioctld,(void *)arg,sizeof(NuScsiTcpIoctlData)) ) {
            DEBUG_ERR0("iSCSI: Cannot copy user Ioctl args\n");
            return -EINVAL;
        }

        if ( ioctld.NumTargets <= 0 ) {
            return -EINVAL;
        }

        rc = NuScsiTcpConnection( &ioctld );

        return rc;
    }

#ifdef DEBUG_TRACE
    if ( cmd == NUSCSITCP_GETTRACE ) {
        if (copy_to_user((void *)arg, &TraceTable[0],
                         (sizeof(NuScsiTcpTraceEntry)*NUSCSITCP_TRACE_COUNT)))
            return -EFAULT;
        return TraceIndex;
    }
#endif

    return -EINVAL;
}

Scsi_Host_Template NuScsiTcpDriverTemplate = NUSCSITCP;

int
init_module( void )
{
#ifndef USE_DYNAMIC_DEVICE
    int timo;
#endif

#if defined(INTERNAL_DRIVER_VERSION) && (INTERNAL_DRIVER_VERSION > 0)
        /* show internal version number */
        printk("iSCSI, version %d.%d.%d.%d (%s)\n",
               (SCSI_TCP_ISCSI_VERSION+1), SCSI_TCP_CISCO_VERSION, SCSI_TCP_DRIVER_VERSION, INTERNAL_DRIVER_VERSION,
               ISCSI_DATE);
#else
        /* released version number (INTERNAL == 0) */
        printk("iSCSI, version %d.%d.%d (%s)\n",
               (SCSI_TCP_ISCSI_VERSION+1), SCSI_TCP_CISCO_VERSION, SCSI_TCP_DRIVER_VERSION,
               ISCSI_DATE);
#endif

#if defined(BUILD_STR)
        /* show a build string if we have one */
        if (NULL != BUILD_STR) {
            printk("iSCSI, %s\n", (char *)BUILD_STR);
        }
#endif

    ControlMajor = register_chrdev( 0, ControlName, &ControlFops );
    if ( ControlMajor < 0 ) {
        printk("iSCSI: failed to register the control device\n");
        return ControlMajor;
    }

#ifdef USE_DYNAMIC_DEVICE
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    NuScsiTcpDriverTemplate.module = THIS_MODULE;
#else
    NuScsiTcpDriverTemplate.module = &__this_module;
#endif
    scsi_register_module( MODULE_SCSI_HA, &NuScsiTcpDriverTemplate );
    if ( ! NuScsiTcpDriverTemplate.present ) {
        DEBUG_ERR0("iSCSI: scsi_register not present\n");
        scsi_unregister_module( MODULE_SCSI_HA, &NuScsiTcpDriverTemplate );
    }
#else
    while( TimerActive == 0 ) {
        if ( signal_pending(current) ) {
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        timo = schedule_timeout( 10 );
    }

    if ( TimerActive ) {
        NuScsiTcpDriverTemplate.module = &__this_module;
        scsi_register_module( MODULE_SCSI_HA, &NuScsiTcpDriverTemplate );
        if ( ! NuScsiTcpDriverTemplate.present ) {
            DEBUG_ERR0("iSCSI: scsi_register not present\n");
            scsi_unregister_module( MODULE_SCSI_HA, &NuScsiTcpDriverTemplate );
        }
    }
#endif
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    init_waitqueue_head(&WaitQ);
#endif

    return 0;
}

void
cleanup_module( void )
{
    int rc;

    DEBUG_INIT0("iSCSI: cleanup\n");
    if ( ControlMajor > 0 ) {
        rc = unregister_chrdev( ControlMajor, ControlName );
        if ( rc ) {
            DEBUG_ERR0("iSCSI: Error trying to unregister control device\n");
        } else {
            ControlMajor = 0;
        }
    }

    if ( NuScsiTcpDriverTemplate.present ) {
        DEBUG_INIT0("iSCSI: SCSI present\n");
        scsi_unregister_module( MODULE_SCSI_HA, &NuScsiTcpDriverTemplate );
        NuScsiTcpDriverTemplate.present = 0;
    }

    return;
}

static int
SetDirection( unsigned char cmd )
{
    switch( cmd ) {
        case TEST_UNIT_READY:
        case    START_STOP:
        case REZERO_UNIT:
        case WRITE_FILEMARKS:
        case SPACE:
        case ERASE:
        case ALLOW_MEDIUM_REMOVAL:
            /* just control commands */
            return NUSCSITCP_CMD_CONTROL;
      case WRITE_6:           case WRITE_10:          case WRITE_12:
      case WRITE_LONG:        case WRITE_SAME:        case WRITE_BUFFER:
      case WRITE_VERIFY:      case WRITE_VERIFY_12:
      case COMPARE:           case COPY:              case COPY_VERIFY:
      case SEARCH_EQUAL:      case SEARCH_HIGH:       case SEARCH_LOW:
      case SEARCH_EQUAL_12:   case SEARCH_HIGH_12:    case SEARCH_LOW_12:
      case FORMAT_UNIT:       case REASSIGN_BLOCKS:   case RESERVE:
      case MODE_SELECT:       case MODE_SELECT_10:    case LOG_SELECT:
      case SEND_DIAGNOSTIC:   case CHANGE_DEFINITION: case UPDATE_BLOCK:
      case SET_WINDOW:        case MEDIUM_SCAN:       case SEND_VOLUME_TAG:
      case 0xea:
            return NUSCSITCP_CMD_WRITE;
      default:
         return NUSCSITCP_CMD_READ;
    }

    return -1;
}

static void
SetTag( Scsi_Cmnd *cmd, struct ScsiTcpScsiCmdHdr *hdr )
{
    if ( cmd->device->tagged_supported ) {
        switch( cmd->tag ) {
            case HEAD_OF_QUEUE_TAG:
                hdr->flags.attr = SCSI_TCP_ATTR_HEAD_OF_QUEUE;
                break;
            case ORDERED_QUEUE_TAG:
                hdr->flags.attr = SCSI_TCP_ATTR_ORDERED;
                break;
            default:
                hdr->flags.attr = SCSI_TCP_ATTR_SIMPLE;
                break;
        }
    }
    return;
}

int
NuScsiTcpXmitCmd( ConnInfoEntry *cie, Scsi_Cmnd *sc, NuScsiTcpCmdEntry *mycmd )
{
    struct msghdr msg;
    struct iovec iov;
    struct ScsiTcpScsiCmdHdr stsch;
    int rc, wlen, directx;

    DEBUG_FLOW4("iSCSI: Cmd 0x%x, Targ %d, Lun %d, Chan %d\n",sc->cmnd[0],
        sc->target,sc->lun,sc->channel);
    DEBUG_FLOW3("       CmdLen 0x%x, ReqLen 0x%x, BufLen 0x%x\n",sc->cmd_len,
        sc->request_bufflen,sc->bufflen);

    wlen = sizeof(stsch);
    memset( &stsch, 0, sizeof(stsch) );
    directx = SetDirection( sc->cmnd[0] );
    if ( directx < 0 ) {
            printk("iSCSI: Unsupported command, 0x%x\n",sc->cmnd[0]);
            return -1;
    }
    if ( directx != NUSCSITCP_CMD_CONTROL ) {
        /* read or write type command */
        if ( directx == NUSCSITCP_CMD_READ ) {
            /* read */
            stsch.flags.read_data = 1;
        }
        stsch.data_length = htonl(sc->request_bufflen);
    }

    SetTag( sc, &stsch );
    stsch.opcode = SCSI_TCP_OP_SCSI_CMD;
    stsch.itt = htonl((unsigned int)mycmd);
    stsch.cmdrn = htonl(cie->CmdRn);
    stsch.expstatrn = htonl(cie->ExpStatRn);
    if ( sc->target >= NumTargets ) {
        DEBUG_ERR1("iSCSI: TargetId too large, 0x%x\n",sc->target);
        return -1;
    }
    if ( NumTargets == 1 ) {
        stsch.lun[1] = sc->lun;
    } else {
        stsch.lun[0] = (TargetIds[sc->target] & 0x3f) | 0x80;
        stsch.lun[1] = (sc->lun & 0x1f);
    }
    memcpy( stsch.scb, sc->cmnd, sc->cmd_len );

    iov.iov_base = &stsch;
    iov.iov_len = sizeof(stsch);
    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    rc = MySendMsg( cie, &msg, wlen );
    if ( rc != wlen ) {
        DEBUG_ERR2("iSCSI: Error XmitCmd %d, %d\n",wlen,rc);
        /* close the connection */
        cie->ConnFlag &= ~CONN_ALIVE;
        return -1;
    }

    NU_TRACE( NUSCSITCP_TRACE_TxCmd, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, sc->request_bufflen );

    return directx;
}

void
NuScsiTcpRecvCmd( struct ScsiTcpScsiRspHdr *stsrh, unsigned char *xbuf )
{
    NuScsiTcpCmdEntry *mycmd;
    Scsi_Cmnd *sc;
    unsigned char *respbuf;
    int senselen;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif
    union {
        unsigned int l;
        unsigned char c[4];
    } ltoc;

    mycmd = (NuScsiTcpCmdEntry *)ntohl(stsrh->itt);
    if ( (mycmd < HostA->cmd_mem) || (mycmd > HostA->cmd_max) || !mycmd->sc ) {
        DEBUG_ERR0("iSCSI: received command response with nothing queued\n");
        return;
    }
    sc = mycmd->sc;

    DEBUG_FLOW3("iSCSI: RxCmd CmdLen %d, Cmd 0x%x, RxLen 0x%x\n",
        sc->cmd_len,sc->cmnd[0],ntohl(stsrh->length));

    if ( stsrh->iscsi_status ) {
        /*
         * If this is not giving the right status, remove this
         * check and let the cmd_status check handle it.
         */
        NU_TRACE( NUSCSITCP_TRACE_RxCmdStatus, sc->lun, sc->cmnd[0], sc->target,
            (unsigned int)mycmd, stsrh->iscsi_status );
        DEBUG_INIT1("iSCSI: RxCmd status 0x%x\n",stsrh->iscsi_status);
        sc->result = DID_NO_CONNECT << 16;
    } else {
        sc->result = stsrh->cmd_status;
        if ( stsrh->cmd_status ) {
            senselen = ntohl(stsrh->length);
            NU_TRACE( NUSCSITCP_TRACE_RxCmdCmdStatus, sc->lun, sc->cmnd[0],
                sc->target, (unsigned int)mycmd, stsrh->cmd_status );
            DEBUG_INIT3("iSCSI: RxCmd (0x%x) bad CmdStatus 0x%x, RxLen 0x%x\n",
                sc->cmnd[0],stsrh->cmd_status,senselen);
            if ( senselen ) {
                printk("iSCSI: Tgt %d, Lun %d, Cmd 0x%x, Sense %02x%02x%02x%02x %02x%02x%02x%02x\n",
                    sc->target,sc->lun,sc->cmnd[0],xbuf[0],xbuf[1],xbuf[2],xbuf[3],
                    xbuf[4],xbuf[5],xbuf[6],xbuf[7]);
                if ( xbuf[7] == 10 ) {
                    printk("      %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x\n",
                        xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                        xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                        xbuf[16],xbuf[17]);
                } else {
                    printk("      %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x\n",
                        xbuf[8],xbuf[9],xbuf[10],xbuf[11],
                        xbuf[12],xbuf[13],xbuf[14],xbuf[15],
                        xbuf[16],xbuf[17],xbuf[18],xbuf[19],
                        xbuf[20],xbuf[21],xbuf[22],xbuf[23]);
                }
                if ( senselen > sizeof(sc->sense_buffer) ) {
                    senselen = sizeof(sc->sense_buffer);
                }
                memcpy( sc->sense_buffer, xbuf, senselen );
            }
        }
    }

    if ( (mycmd->direction == NUSCSITCP_CMD_READ) &&
         (mycmd->rxdata != sc->request_bufflen) ) {
        mycmd->rxrsp = 1;
        if ( stsrh->flags.overflow || stsrh->flags.underflow ) {
            mycmd->rxovrflw = stsrh->flags.overflow;
            mycmd->rxundflw = stsrh->flags.underflow;
            DEBUG_INIT3("iSCSI: Cmd 0x%x, OvrFlw %d, UndFlw %d\n",sc->cmnd[0],
                stsrh->flags.overflow,stsrh->flags.underflow);
            DEBUG_INIT3("           RxData 0x%x, ResCnt 0x%x, ReqLen 0x%x\n",
                mycmd->rxdata,ntohl(stsrh->residual_count),sc->request_bufflen);
            if ( stsrh->flags.underflow ) {
                NU_TRACE( NUSCSITCP_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                    sc->target, (unsigned int)mycmd, ntohl(stsrh->residual_count) );
#if LINUX_VERSION_CODE >= LINUXVERSION(2,3,19)
                sc->resid = ntohl(stsrh->residual_count);
#else
                if ( mycmd->rxdata < sc->underflow ) {
                    sc->result = DID_ERROR;
                }
#endif
            } else {
                NU_TRACE( NUSCSITCP_TRACE_RxCmdFlow, sc->lun, sc->cmnd[0],
                    sc->target, (unsigned int)mycmd, 0 );
            }
        } else {
            /* All the data did not arrive */
            sc->result = DID_ERROR;
        }
    }

    /* command is done and everything we are going to get has arrived */

    NU_TRACE( NUSCSITCP_TRACE_RxCmd, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, mycmd->rxdata );

    if ( mycmd->cmdaborted ) {
        DEBUG_INIT0("iSCSI: RxCmd CmdAborted, not done\n");
        mycmd->abortdone = 1;
        return;
    }

    FreeMyCmd( mycmd );
    sc->host_scribble = NULL;

    if ( sc->cmnd[0] == TEST_UNIT_READY ) {
        if ( stsrh->cmd_status == 0 ) {
            if ( (TargetLuns[sc->target] & (1 << sc->lun)) == 0 ) {
                TargetLuns[sc->target] |= (1 << sc->lun);
                ltoc.l = ConnInfo[sc->target]->RemoteIpAddr;
                printk("iSCSI: Tgt 0x%x, Lun 0x%x, Remote %u.%u.%u.%u, TgtId 0x%x, %s\n",
                       sc->target,sc->lun,ltoc.c[0],ltoc.c[1],ltoc.c[2],ltoc.c[3], /* sd_devname(DEVICE_NR(sc->request.rq_dev)), */
                       TargetIds[sc->target],TargetNames[sc->target]);
                /* FIXME: display & record the Linux device name */
            }
        }
        respbuf = sc->request_buffer;
        memset( respbuf, 0, sc->request_bufflen );
    }

    NU_TRACE( NUSCSITCP_TRACE_CmdDone, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, 0 );
    spin_lock_irqsave( &io_request_lock, cpu_flags );
    sc->scsi_done( sc );
    spin_unlock_irqrestore( &io_request_lock, cpu_flags );

    return;
}

void
NuScsiTcpQueueRttData( struct ScsiTcpRttHdr *strh )
{
    Scsi_Cmnd *sc;
    NuScsiTcpCmdEntry *mycmd;

    DEBUG_FLOW3("iSCSI: RTT Len 0x%x, Off 0x%x, Itt 0x%x\n",
        ntohl(strh->data_length),ntohl(strh->data_offset),ntohl(strh->itt));

    mycmd = (NuScsiTcpCmdEntry *)ntohl(strh->itt);
    if ( (mycmd < HostA->cmd_mem) || (mycmd > HostA->cmd_max) || !mycmd->sc ) {
        DEBUG_ERR0("iSCSI: Rtt with nothing queued\n");
        return;
    }

    if ( strh->data_offset ) {
        mycmd->data_offset = strh->data_offset;
        DEBUG_ERR1("iSCSI: Rtt data with offset -- 0x%x\n",
            ntohl(mycmd->data_offset));
        return;
    }

    sc = mycmd->sc;
    if ( ntohl(strh->data_length) != sc->request_bufflen ) {
        DEBUG_ERR2("iSCSI: Rtt data bad length -- 0x%x, 0x%x\n",
            ntohl(strh->data_length), sc->request_bufflen);
        return;
    }

    mycmd->ttt = strh->ttt;
    mycmd->senddata = 1;
    NU_TRACE( NUSCSITCP_TRACE_RttData, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, mycmd->ttt );

    wake_up_interruptible( &WaitQ );

    return;
}

void
NuScsiTcpXmitRttData( ConnInfoEntry *cie, NuScsiTcpCmdEntry *mycmd )
{
    struct msghdr msg;
    struct ScsiTcpDataHdr stdh;
    Scsi_Cmnd *sc;
    struct scatterlist *sglist;
    int length, wlen, rc, iovn;

    sc = mycmd->sc;
    length = sc->request_bufflen;

    memset( &stdh, 0, sizeof(stdh) );
    stdh.opcode = SCSI_TCP_OP_SCSI_DATA;
    stdh.length = htonl(length);
    stdh.ttt = mycmd->ttt;
    stdh.offset = mycmd->data_offset;
    stdh.itt = htonl((unsigned int)mycmd);
    stdh.expstatrn = htonl(cie->ExpStatRn);

    wlen = sizeof(stdh) + length;

    cie->TxIov[0].iov_base = &stdh;
    cie->TxIov[0].iov_len = sizeof(stdh);
    if ( sc->use_sg ) {
        iovn = 1;
        sglist = (struct scatterlist *)sc->request_buffer;
        rc = 0;
        while( rc < sc->use_sg ) {
            cie->TxIov[iovn].iov_base = sglist->address;
            cie->TxIov[iovn].iov_len = sglist->length;
            iovn += 1;
            rc += 1;
            sglist += 1;
        }
    } else {
        cie->TxIov[1].iov_base = sc->request_buffer;
        cie->TxIov[1].iov_len = length;
        iovn = 2;
    }

    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &cie->TxIov[0];
    msg.msg_iovlen = iovn;
    /* msg.msg_flags = MSG_DONTWAIT; */

    rc = MySendMsg( cie, &msg, wlen );
    if ( rc != wlen ) {
        DEBUG_ERR2("iSCSI: Error XmitRttData %d, %d\n",wlen,rc);
        /* close the connection */
        cie->ConnFlag &= ~CONN_ALIVE;
        return;
    }

    NU_TRACE( NUSCSITCP_TRACE_TxData, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, mycmd->ttt );

    return;
}

void
NuScsiTcpXmitImmData( ConnInfoEntry *cie, NuScsiTcpCmdEntry *mycmd )
{
    struct msghdr msg;
    struct ScsiTcpDataHdr stdh;
    Scsi_Cmnd *sc;
    struct scatterlist *sglist;
    int wlen, rc, iovn, length, iovlen;

    sc = mycmd->sc;

    length = sc->request_bufflen;

    memset( &stdh, 0, sizeof(stdh) );
    stdh.opcode = SCSI_TCP_OP_SCSI_DATA;
    stdh.length = htonl(length);
    stdh.ttt = htonl((unsigned int)mycmd);
    stdh.itt = stdh.ttt;
    stdh.expstatrn = htonl(cie->ExpStatRn);

    wlen = sizeof(stdh) + length;

    cie->TxIov[0].iov_base = &stdh;
    cie->TxIov[0].iov_len = sizeof(stdh);
    if ( sc->use_sg ) {
        iovn = 1;
        sglist = (struct scatterlist *)sc->request_buffer;
        iovlen = sizeof(stdh);
        rc = 0;
        while( rc < sc->use_sg ) {
            cie->TxIov[iovn].iov_base = sglist->address;
            if ( (sglist->length+iovlen) > wlen ) {
                cie->TxIov[iovn].iov_len = wlen - iovlen;
                iovn += 1;
                break;
            }
            iovlen += sglist->length;
            cie->TxIov[iovn].iov_len = sglist->length;
            iovn += 1;
            rc += 1;
            sglist += 1;
        }
    } else {
        cie->TxIov[1].iov_base = sc->request_buffer;
        cie->TxIov[1].iov_len = length;
        iovn = 2;
    }

    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &cie->TxIov[0];
    msg.msg_iovlen = iovn;
    /* msg.msg_flags = MSG_DONTWAIT; */

    rc = MySendMsg( cie, &msg, wlen );
    if ( rc != wlen ) {
        DEBUG_ERR2("iSCSI: Error XmitImmData %d, %d\n",wlen,rc);
        /* close the connection */
        cie->ConnFlag &= ~CONN_ALIVE;
        return;
    }

    NU_TRACE( NUSCSITCP_TRACE_TxData, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, mycmd->ttt );

    return;
}

void
NuScsiTcpFillScsiCmd( ConnInfoEntry *cie, struct ScsiTcpDataRspHdr *stdrh )
{
    NuScsiTcpCmdEntry *mycmd;
    Scsi_Cmnd *sc;
    struct scatterlist *sglist;
    int cmdlen, iovn, rc;
    struct msghdr msg;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif

    mycmd = (NuScsiTcpCmdEntry *)ntohl(stdrh->itt);
    if ( (mycmd < HostA->cmd_mem) || (mycmd > HostA->cmd_max) || !mycmd->sc ) {
        DEBUG_ERR0("iSCSI: FillScsiCmd with no command\n");
        return;
    }

    sc = mycmd->sc;
    cmdlen = ntohl( stdrh->length );
    if ( sc->use_sg ) {
        iovn = 0;
        sglist = (struct scatterlist *)sc->request_buffer;
        while( iovn < sc->use_sg ) {
            cie->RxIov[iovn].iov_base = sglist->address;
            cie->RxIov[iovn].iov_len = sglist->length;
            iovn += 1;
            sglist += 1;
        }
    } else {
        cie->RxIov[0].iov_base = sc->request_buffer;
        cie->RxIov[0].iov_len = cmdlen;
        iovn = 1;
    }
    memset( &msg, 0, sizeof(struct msghdr) );
    msg.msg_iov = &cie->RxIov[0];
    msg.msg_iovlen = iovn;

    rc = MyRecvMsg( cie, &msg, cmdlen );
    if ( rc <= 0 ) {
        DEBUG_ERR1("iSCSI: RxData bad rx response, %d\n",rc);
        return;
    }
    if ( rc != cmdlen ) {
        DEBUG_ERR2("iSCSI: RxData not full 0x%x, 0x%x\n",cmdlen,rc);
    }

    mycmd->rxdata += rc;

    NU_TRACE( NUSCSITCP_TRACE_RxData, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, cmdlen );

    if ( mycmd->rxrsp && (mycmd->rxdata == sc->request_bufflen) ) {
        if ( mycmd->cmdaborted ) {
            DEBUG_ERR0("iSCSI: FillCmd CmdAborted, not done\n");
            mycmd->abortdone = 1;
            return;
        }
        FreeMyCmd( mycmd );
        sc->host_scribble = NULL;
        NU_TRACE( NUSCSITCP_TRACE_CmdDone, sc->lun, sc->cmnd[0], sc->target,
            (unsigned int)mycmd, 0 );
        if ( sc->result == 0 ) {
            spin_lock_irqsave( &io_request_lock, cpu_flags );
            sc->scsi_done( sc );
            spin_unlock_irqrestore( &io_request_lock, cpu_flags );
        }
    }

    return;
}

int
NuScsiTcpTimer( void )
{
    int rc;
    unsigned long schold_timeout;
    Scsi_Cmnd *sc, *scq, *schold_head, *schold_tail;
    ConnInfoEntry *cie;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif

    DEBUG_INIT0("iSCSI: Timer\n");

    TimerActive = 1;

    while( ! HostA ) {
        if ( signal_pending(current) ) {
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        rc = schedule_timeout( 10 );
    }

    schold_head = schold_tail = NULL;
    schold_timeout = 0;

    while( HostA ) {
        interruptible_sleep_on_timeout( &WaitQ, 10 );
        if ( signal_pending(current) ) {
            break;
        }

        if ( HostA->q_head ) {
            scq = HostA->q_head;
            HostA->q_head = NULL;
            while ( scq ) {
                sc = scq;
                scq = (Scsi_Cmnd *)sc->host_scribble;
                sc->host_scribble = NULL;
                if ( sc->result == 0 ) {
                    cie = ConnInfo[sc->target];
                    if ( cie && (schold_head == NULL) ) {
                        down( &cie->ConnLock );
                        if ( cie->ConnScHead ) {
                            cie->ConnScTail->host_scribble = (char *)sc;
                        } else {
                            cie->ConnScHead = sc;
                        }
                        cie->ConnScTail = sc;
                        up( &cie->ConnLock );
                        cie->ConnFlag |= CONN_WAKE;
                    } else {
                        if ( schold_head ) {
                            schold_tail->host_scribble = (char *)sc;
                        } else {
                            schold_head = sc;
                            schold_timeout = jiffies + (5*HZ);
                        }
                        schold_tail = sc;
                    }
                }
            }
        } else {
            cie = ConnInfoHead;
            while ( cie ) { 
                if ( cie->Jiffies && (cie->Jiffies < jiffies) ) {
                    if ( (cie->ConnFlag & CONN_PINGRSP) ||
                         (cie->ConnFlag & CONN_PING) ) {
                        if ( cie->ConnFlag & CONN_DEAD ) {
                            printk("iSCSI: Connection, 0x%x, appears dead\n",
                                   htonl(cie->RemoteIpAddr));
                            if ( cie->TaskPid ) {
                                DEBUG_INIT1("iSCSI: killing TaskPid %d\n", cie->TaskPid);
                                kill_proc(cie->TaskPid, SIGKILL, 1);
                            }
                            if ( cie->ConnPid ) {
                                DEBUG_INIT1("iSCSI: killing ConnPid %d\n", cie->ConnPid);
                                kill_proc(cie->ConnPid, SIGPIPE, 1);
                            }
                            DEBUG_INIT1("iSCSI: connection, 0x%x, appears dead\n",
                                        htonl(cie->RemoteIpAddr));
                            cie->ConnFlag &= ~(CONN_ALIVE|CONN_PINGRSP|CONN_DEAD);
                            cie->Jiffies = 0;
                        } else {
                            DEBUG_INIT1("iSCSI: no ping response from connection, 0x%x\n",
                                htonl(cie->RemoteIpAddr));
                            cie->ConnFlag |= (CONN_PING | CONN_DEAD);
                            cie->Jiffies = jiffies + NUSCSITCP_TIMEOUT;
                        }
                    } else {
                        DEBUG_INIT1("iSCSI: connection, 0x%x, has command pending\n",
                            htonl(cie->RemoteIpAddr));
                        cie->ConnFlag |= CONN_PING;
                        cie->Jiffies = jiffies + NUSCSITCP_TIMEOUT;
                    }
                } else {
                    if ( (jiffies-cie->PingTimeout) >= (60*HZ) ) {
                        if ( (cie->RxBytes == 0) && (cie->ConnFlag & CONN_ALIVE) ) {
                            cie->ConnFlag |= CONN_PING;
                        }
                        if ( cie == ConnInfoHead ) {
                            DEBUG_INIT3("iSCSI: 0x%08x - RxBytes %d, TxBytes %d\n",
                                htonl(cie->RemoteIpAddr),cie->RxBytes,cie->TxBytes);
                        } else {
                            DEBUG_INIT3("         : 0x%08x - RxBytes %d, TxBytes %d\n",
                                htonl(cie->RemoteIpAddr),cie->RxBytes,cie->TxBytes);
                        }
                        cie->RxBytes = cie->TxBytes = 0;
                        cie->PingTimeout = jiffies;
                    }
                }

                cie = cie->Next;
            }

            scq = schold_head;
            schold_head = NULL;
            while( scq ) {
                sc = scq;
                scq = (Scsi_Cmnd *)sc->host_scribble;
                sc->host_scribble = NULL;
                cie = ConnInfo[sc->target];
                if ( cie ) {
                    down( &cie->ConnLock );
                    if ( cie->ConnScHead ) {
                        cie->ConnScTail->host_scribble = (char *)sc;
                    } else {
                        cie->ConnScHead = sc;
                    }
                    cie->ConnScTail = sc;
                    up( &cie->ConnLock );
                    cie->ConnFlag |= CONN_WAKE;
                } else {
                    if ( jiffies >= schold_timeout ) {
                        sc->result = DID_NO_CONNECT << 16;
                        spin_lock_irqsave( &io_request_lock, cpu_flags );
                        sc->scsi_done( sc );
                        spin_unlock_irqrestore( &io_request_lock, cpu_flags );
                    } else {
                        if ( schold_head ) {
                            schold_tail->host_scribble = (char *)sc;
                        } else {
                            schold_head = sc;
                        }
                        schold_tail = sc;
                    }
                }
            }
        }

        for( rc=0; rc<NumTargets; rc++ ) {
            if ( ConnInfo[rc] ) {
                if ( ConnInfo[rc]->ConnFlag & CONN_WAKE ) {
                    wake_up_interruptible( &ConnInfo[rc]->ConnWaitQ );
                    ConnInfo[rc]->ConnFlag &= ~CONN_WAKE;
                }
            }
        }
    }

    TimerActive = 0;
    NumTargets = 0;

    current->state = TASK_RUNNING;
    DEBUG_INIT0("iSCSI: Timer leaving kernel\n");

    return 0;
}

int
NuScsiTcpConnection( NuScsiTcpIoctlData *ioctld )
{
    int i, rc, xlen, ti, ci;
    int status = 0;
    struct socket *sock;
    struct msghdr msg;
    struct iovec iov;
    struct inode *ctrlinode;
    struct file *ctrlfile;
    struct ScsiTcpHdr sth;
    unsigned char *rxbuf;
    ConnInfoEntry *cie;
    NuScsiTcpCmdEntry *mycmd;
    Scsi_Cmnd *sc;
    mm_segment_t fs;

    /* get the socket passed in for the connection */
    ctrlfile = fget( ioctld->TcpSock );
    if ( ! ctrlfile ) {
        return -EINVAL;
    }

    ctrlinode = ctrlfile->f_dentry->d_inode;
    if ( !ctrlinode || !ctrlinode->i_sock ) {
        fput( ctrlfile );
        return -EINVAL;
    }

    sock = &ctrlinode->u.socket_i;
    if ( sock->file != ctrlfile ) {
        fput( ctrlfile );
        return -EINVAL;
    }

    sock->sk->allocation = GFP_ATOMIC;

    /* see if we already have a connection info entry for this connection */
    cie = ConnInfoHead;
    ti = 0;
    while( cie ) {
        if ( cie->RemoteIpAddr == ioctld->RemoteIpAddr ) {
            /* FIXME: must also check target */
            /*
             * connection must have been re-started
             */
            if ( (cie->NumTargets != ioctld->NumTargets) && cie->Next ) {
                DEBUG_ERR2(
                    "NuScsiTcp: reconnection with different target count, %d - %d\n",
                    cie->NumTargets,ioctld->NumTargets);
                fput( ctrlfile );
                return -EINVAL;
            }
            break;
        }
        ti += cie->NumTargets;
        cie = cie->Next;
    }

    if ( cie ) {
        printk("iSCSI: Restarted connection, 0x%x\n", htonl(ioctld->RemoteIpAddr));
        for( ci=0; ci<ioctld->NumTargets; ci++,ti++ ) {
            TargetIds[ti] = ioctld->TargetIds[ci];
            ConnInfo[ti] = cie;
            memcpy(TargetNames[ti], ioctld->TargetNames[ci], 16);
        }
    } else {
        /*
         * must be a new connection
         */
        printk("iSCSI: New connection, 0x%x\n", htonl(ioctld->RemoteIpAddr));

        if ( (ioctld->NumTargets+NumTargets) > NUSCSITCP_MAX_TARGETS ) {
            DEBUG_ERR0("iSCSI: Too many targets\n");
            fput( ctrlfile );
            return -EINVAL;
        }
        cie = (ConnInfoEntry *)kmalloc( sizeof(ConnInfoEntry), GFP_KERNEL );
        if ( ! cie ) {
            DEBUG_ERR0("iSCSI: Cannot allocate new connection entry\n");
            fput( ctrlfile );
            return -ENOMEM;
        }
        if ( ConnInfoHead ) {
            ConnInfoTail->Next = cie;
        } else {
            ConnInfoHead = cie;
        }
        ConnInfoTail = cie;
        memset( cie, 0, sizeof(ConnInfoEntry) );
        cie->NumTargets = ioctld->NumTargets;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
        /* see include/linux/semaphore.h */
        init_MUTEX(&cie->ConnLock);
#else
        cie->ConnLock = MUTEX;
#endif
        cie->RemoteIpAddr = ioctld->RemoteIpAddr;
        cie->ConnI = NumConnections;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
        init_waitqueue_head(&cie->ConnWaitQ);
#endif
        for( i=0; i<ioctld->NumTargets; i++ ) {
            TargetIds[(NumTargets+i)] = ioctld->TargetIds[i];
            ConnInfo[(NumTargets+i)] = cie;
            memcpy( TargetNames[(NumTargets+i)], ioctld->TargetNames[i], 16 );
        }
        NumTargets += ioctld->NumTargets;
        DEBUG_INIT3("iSCSI: Target count %d: %d/%d\n",NumConnections,
            ioctld->NumTargets,NumTargets);
        NumConnections += 1;
    }

    cie->TcpSock = sock;
    cie->RxBytes = 0;
    cie->TxBytes = 0;
    cie->CmdRn = 1;
    cie->ExpStatRn = 0;
    cie->PingCount = 0;
    cie->UseRtt = ioctld->UseRtt;
    cie->ConnFlag = CONN_ALIVE;
    cie->PingTimeout = jiffies;

    sprintf(current->comm,"NuConnection%d",cie->ConnI);

#ifdef USE_DYNAMIC_DEVICE
    while( ! TimerActive ) {
        if ( signal_pending(current) ) {
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        rc = schedule_timeout( 10 );
    }
#endif

    cie->ConnPid = current->pid;
#ifdef ISCSI_CLONE
    cie->TaskPid = kernel_thread( NuScsiTcpConnTask, (void *)cie, 0 );
#else
    cie->TaskPid = kernel_thread( NuScsiTcpConnTask, (void *)cie,
        (CLONE_VM|CLONE_FS|CLONE_FILES|CLONE_SIGHAND) );
#endif

    while( ! HostA ) {
        if ( signal_pending(current) ) {
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        rc = schedule_timeout( 10 );
    }

    while( HostA && TimerActive && (cie->ConnFlag & CONN_ALIVE) ) {
        if ( signal_pending(current) ) {
            break;
        }

        /* check for anything to read on socket */
        iov.iov_base = &sth;
        iov.iov_len = sizeof(sth);
        memset( &msg, 0, sizeof(struct msghdr) );
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        rc = MyRecvMsg( cie, &msg, sizeof(sth) );
        if ( rc == sizeof(sth) ) {
            /* received something */
            xlen = ntohl( sth.length );
            if ( xlen && (sth.opcode != SCSI_TCP_OP_SCSI_DATA_RSP) ) {
                if ( xlen > NUSCSITCP_RXCTRL_SIZE ) {
                    DEBUG_ERR1("iSCSI: packet length too large, %d\n", xlen);
                    continue;
                }
                rxbuf = cie->RxBuf;
                iov.iov_base = rxbuf;
                iov.iov_len = xlen;
                memset( &msg, 0, sizeof(struct msghdr) );
                msg.msg_iov = &iov;
                msg.msg_iovlen = 1;
                rc = MyRecvMsg( cie, &msg, xlen );
                if ( rc != xlen ) {
                    DEBUG_ERR2("iSCSI: packet bad append, %d, %d\n",xlen,rc);
                    continue;
                }
            } else {
                rxbuf = NULL;
            }

            switch( sth.opcode ) {
                case SCSI_TCP_OP_NOOP_IN:
                    NuScsiTcpXmitNop( cie, (struct ScsiTcpNopInHdr *)&sth );
                    break;
                case SCSI_TCP_OP_SCSI_RSP:
                    cie->ExpStatRn = ntohl(((struct ScsiTcpScsiRspHdr *)&sth)->statrn)+1;
                    NuScsiTcpRecvCmd( (struct ScsiTcpScsiRspHdr *)&sth, rxbuf );
                    break;
                case SCSI_TCP_OP_SCSI_TASK_MGT_RSP:
                    cie->ExpStatRn = ntohl(((struct ScsiTcpScsiTaskMgtRspHdr *)&sth)->statrn)+1;
                    NuScsiTcpMgmtResp( (struct ScsiTcpScsiTaskMgtRspHdr *)&sth );
                    break;
                case SCSI_TCP_OP_RTT_RSP:
                    NuScsiTcpQueueRttData( (struct ScsiTcpRttHdr *)&sth );
                    break;
                case SCSI_TCP_OP_SCSI_DATA_RSP:
                    cie->ExpStatRn = ntohl(((struct ScsiTcpDataRspHdr *)&sth)->statrn)+1;
                    NuScsiTcpFillScsiCmd( cie, (struct ScsiTcpDataRspHdr *)&sth );
                    break;
                case SCSI_TCP_OP_PING_RSP:
                    NU_TRACE( NUSCSITCP_TRACE_RxPing, 0, 0, 0, 0, 0 );
                    DEBUG_FLOW0("iSCSI: RxPing\n");
                    cie->ConnFlag &= ~(CONN_PINGRSP | CONN_DEAD);
                    cie->ExpStatRn = ntohl(((struct ScsiTcpPingRspHdr *)&sth)->statrn)+1;
                    break;
                case SCSI_TCP_OP_ASYNC_EVENT:
                    cie->ExpStatRn = ntohl(((struct ScsiTcpAsyncEvtHdr *)&sth)->statrn)+1;
                    NuScsiTcpAsyncEvent( (struct ScsiTcpAsyncEvtHdr *)&sth );
                    break;
                default:
                    DEBUG_ERR1("iSCSI: unexpected RxCmd opcode, 0x%x\n",
                               sth.opcode);
                    goto ConnectionDone;
            }
        } else {
            if ( rc != -EAGAIN ) {
                if ( (rc == -ECONNRESET) || (rc == 0) ) {
                    DEBUG_ERR1( "iSCSI: Connection, 0x%x, reset.\n",
                        htonl(cie->RemoteIpAddr));
                } else {
                    if ( rc == -ERESTARTSYS ) {
                        DEBUG_ERR1( "iSCSI: Connection, 0x%x, terminated.\n",
                            htonl(cie->RemoteIpAddr));
                    } else {
                        DEBUG_ERR2(
                            "iSCSI: Unexpected read status on connection 0x%x, %d\n",
                            htonl(cie->RemoteIpAddr),rc);
                    }
                }
                goto ConnectionDone;
            }
        }
    }

ConnectionDone:
    printk("iSCSI: Connection, 0x%x, going down\n", htonl(cie->RemoteIpAddr));
    cie->ConnFlag &= ~CONN_ALIVE;

    /* indicate that we're already going down, so that we don't get killed */
    cie->ConnPid = 0;

    fs = get_fs();     /* save previous value */
    set_fs (get_ds()); /* use kernel limit */
    rc = waitpid(cie->TaskPid, &status, __WCLONE|WNOHANG);
    set_fs(fs); /* restore before returning to user space */
    
    if (rc == cie->TaskPid) {
        cie->TaskPid = 0;
    }
    else {
        /* if we can't reap it, kill it, and try again later */
        DEBUG_INIT1("iSCSI: connection done, killing TaskPid %d\n", cie->TaskPid);
        kill_proc(cie->TaskPid, SIGKILL, 1);
    }

    /* abort all commands for this connection */
    mycmd = cie->CmdHead;
    while( mycmd ) {
        sc = mycmd->sc;
        printk("iSCSI: Abort command 0x%x, sc 0x%x\n",
            (unsigned int)mycmd,(unsigned int)sc);
        mycmd->sc = NULL;
        FreeMyCmd( mycmd );
        mycmd = cie->CmdHead;
        if ( sc && sc->result == 0 ) {
            sc->host_scribble = NULL;
            sc->result = DID_SOFT_ERROR << 16;
            sc->scsi_done( sc );
        }
    }

    /* disconnect targets */
    for( i=0; i<NumTargets; i++ ) {
        if ( ConnInfo[i] == cie ) {
            ConnInfo[i] = NULL;
        }
    }
    fput( ctrlfile );

    /* wait for the ConnTask to exit. We need a better way of doing
     * this. Perhaps forking in the daemon and using 2 separate ioctls
     * for connection read and write, rather than using
     * kernel_thread() and waitpid() from within the kernel module.
     * On the other hand, if we want to get away from ioctl()s that
     * block for long periods of time, we need to use kernel threads.
     */
    if (cie->TaskPid) {
        DEBUG_INIT1("iSCSI: waiting for TaskPid %d to exit\n", cie->TaskPid);
        fs = get_fs();     /* save previous value */
        set_fs (get_ds()); /* use kernel limit */
        rc = waitpid(cie->TaskPid, &status, __WCLONE);
        set_fs(fs); /* restore before returning to user space */
        if (rc > 0) {
            DEBUG_INIT2("iSCSI: reaped TaskPid %d, status 0x%x\n", cie->TaskPid, status);
            cie->TaskPid = 0;
        }
        else {
            printk("iSCSI: failed to reap TaskPid %d, rc %d, errno %d\n", cie->TaskPid, rc, errno);
        }
    }

    current->state = TASK_RUNNING;
    DEBUG_INIT1("iSCSI: connection, 0x%x, leaving kernel\n",
                ntohl(ioctld->RemoteIpAddr));

    return 0;
}

int
NuScsiTcpXmitMgmtMsg( NuScsiTcpCmdEntry *mycmd, Scsi_Cmnd *sc )
{
    struct msghdr msg;
    struct iovec iov;
    int rc;
    ConnInfoEntry *cie;
    struct ScsiTcpScsiTaskMgtHdr ststmh;

    cie = ConnInfo[sc->target];
    if ( ! cie ) {
        return 1;
    }

    memset( &ststmh, 0, sizeof(ststmh) );
    ststmh.opcode = SCSI_TCP_OP_SCSI_TASK_MGT_MSG;
    ststmh.itt = htonl( (unsigned int)mycmd );
    ststmh.rtt = ststmh.itt;
    if ( NumTargets == 1 ) {
        ststmh.lun[1] = sc->lun;
    } else {
        ststmh.lun[0] = (TargetIds[sc->target] & 0x3f) | 0x80;
        ststmh.lun[1] = (sc->lun & 0x1f);
    }
    ststmh.function = SCSI_TCP_TM_FUNC_ABORT_TASK;
    ststmh.expstatrn = htonl(cie->ExpStatRn);

    iov.iov_base = &ststmh;
    iov.iov_len = sizeof(ststmh);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    rc = MySendMsg( cie, &msg, sizeof(ststmh) );
    if ( rc != sizeof(ststmh) ) {
        DEBUG_ERR1("iSCSI: XmitMsg error %d\n",rc);
        /* close the connection */
        cie->ConnFlag &= ~CONN_ALIVE;
        return 1;
    }

    return 0;
}

void
NuScsiTcpMgmtResp( struct ScsiTcpScsiTaskMgtRspHdr *ststmrh )
{
    NuScsiTcpCmdEntry *mycmd;
    Scsi_Cmnd *sc;

    mycmd = (NuScsiTcpCmdEntry *)ntohl(ststmrh->itt);
    if ( (mycmd < HostA->cmd_mem) || (mycmd > HostA->cmd_max) ) {
        DEBUG_ERR0("iSCSI: received mgmt response with nothing queued\n");
        return;
    }
    if ( ststmrh->response ) {
        DEBUG_ERR0("iSCSI: mgmt response rejected\n");
    }
    DEBUG_ERR2("iSCSI: MgmtResp aborted command, 0x%x -- 0x%x\n",
        (unsigned int)mycmd, ststmrh->response);
    sc = mycmd->sc;
    NU_TRACE( NUSCSITCP_TRACE_RxAbort, sc->lun, sc->cmnd[0], sc->target,
        (unsigned int)mycmd, ststmrh->response );
    mycmd->abortdone = 1;

    return;
}

void
NuScsiTcpAsyncEvent( struct ScsiTcpAsyncEvtHdr *staeh )
{
    DEBUG_ERR2("iSCSI: AsyncEvent scsiEvtInd 0x%x, iscsiEvtInd 0x%x\n",
        staeh->scsi_event_ind, staeh->iscsi_event_ind);
#if 0
    NU_TRACE( NUSCSITCP_TRACE_RxAsnEv, 0, staeh->scsi_event_ind,
        staeh->iscsi_event_ind, staeh->length, 0);
#endif
    return;
}

void
NuScsiTcpXmitPing( ConnInfoEntry *cie )
{
    struct ScsiTcpPingHdr stph;
    struct msghdr msg;
    struct iovec iov;
    int rc;

    memset( &stph, 0, sizeof(stph) );
    stph.opcode = SCSI_TCP_OP_PING_CMD;
    stph.itt = htonl( cie->PingCount );
    stph.expstatrn = htonl(cie->ExpStatRn);
    cie->PingCount += 1;

    iov.iov_base = &stph;
    iov.iov_len = sizeof(stph);
    memset( &msg, 0, sizeof(msg) );
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    rc = MySendMsg( cie, &msg, sizeof(stph) );
    if ( rc != sizeof(stph) ) {
        DEBUG_ERR1("iSCSI: XmitPing error %d\n",rc);
        /* close the connection */
        cie->ConnFlag &= ~CONN_ALIVE;
    }

    NU_TRACE( NUSCSITCP_TRACE_TxPing, 0, 0, 0, 0, cie->RemoteIpAddr );

    return;
}

static NuScsiTcpCmdEntry *
AllocMyCmd( NuScsiTcpHostAdapter *ha, ConnInfoEntry *cie )
{
    NuScsiTcpCmdEntry *ce;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif

    spin_lock_irqsave( &NuScsiTcpSpinLock, cpu_flags );
    ce = ha->free_cmd_head;
    if ( ce ) {
        ha->free_cmd_head = ce->next;
        memset( ce, 0, sizeof(NuScsiTcpCmdEntry) );
        ActiveCmds += 1;
        if ( cie->CmdHead ) {
            cie->CmdTail->next = ce;
            ce->prev = cie->CmdTail;
        } else {
            cie->CmdHead = ce;
            cie->Jiffies = jiffies + NUSCSITCP_TIMEOUT;
        }
        cie->CmdTail = ce;
        ce->jiffies = jiffies;
        ce->cie = cie;
    }
    spin_unlock_irqrestore( &NuScsiTcpSpinLock, cpu_flags );

    return ce;
}

static void
FreeMyCmd( NuScsiTcpCmdEntry *ce )
{
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif
    ConnInfoEntry *cie;

    if ( ! HostA ) {
        DEBUG_ERR0("iSCSI: FreeMyCmd HostA empty\n");
        return;
    }

    cie = ce->cie;
    if ( cie ) {
        spin_lock_irqsave( &NuScsiTcpSpinLock, cpu_flags );
        if ( ce->prev ) {
            ce->prev->next = ce->next;
            if ( ce->next ) {
                ce->next->prev = ce->prev;
            } else {
                cie->CmdTail = ce->prev;
            }
        } else {
            if ( ce->next ) {
                cie->CmdHead = ce->next;
                cie->Jiffies = ce->next->jiffies + NUSCSITCP_TIMEOUT;
            } else {
                cie->CmdHead = NULL;
                cie->Jiffies = 0;
            }
            if ( ce->next ) {
                ce->next->prev = NULL;
            }
        }

        ce->next = NULL;
        if ( HostA->free_cmd_head ) {
            HostA->free_cmd_tail->next = ce;
        } else {
            HostA->free_cmd_head = ce;
        }
        HostA->free_cmd_tail = ce;
        ce->sc = NULL;
        ActiveCmds -= 1;
        spin_unlock_irqrestore( &NuScsiTcpSpinLock, cpu_flags );
    } else {
        DEBUG_ERR1("iSCSI: free cmd not on Q, 0x%x\n",(int)ce);
    }

    return;
}

#ifdef DEBUG_TRACE
void
NuScsiTcpFillTrace( unsigned char flag,
                    unsigned char lun,
                    unsigned char cmd,
                    unsigned char target,
                    unsigned int itt,
                    unsigned int data )
{
    NuScsiTcpTraceEntry *te;
    unsigned int cpu_flags;

    spin_lock_irqsave( &NuScsiTcpSpinLock, cpu_flags );
    te = &TraceTable[TraceIndex];
    TraceIndex += 1;
    if ( TraceIndex >= NUSCSITCP_TRACE_COUNT ) {
        TraceIndex = 0;
    }
    spin_unlock_irqrestore( &NuScsiTcpSpinLock, cpu_flags );

    te->lun = lun;
    te->cmd = cmd;
    te->itt = itt;
    te->flag = flag;
    te->data = data;
    te->target = target;
    te->activecmds = ActiveCmds;
    te->jiffies = jiffies;

    return;
}
#endif

void
NuScsiTcpXmitQdCmds( ConnInfoEntry *cie, Scsi_Cmnd *scq )
{
    int rc;
    Scsi_Cmnd *sc;
    NuScsiTcpCmdEntry *mycmd;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif

    while( scq ) {
        sc = scq;
        scq = (Scsi_Cmnd *)sc->host_scribble;

        if ( sc->result == (DID_ABORT<<16) ) {
            /* already aborted */
            DEBUG_ERR1("iSCSI: Command about to run, but already aborted, 0x%x\n",
                (int)sc);
            continue;
        }

        if ( ! (cie->ConnFlag & CONN_ALIVE) ) {
            sc->result = DID_NO_CONNECT << 16;
            spin_lock_irqsave( &io_request_lock, cpu_flags );
            sc->host_scribble = NULL;
            sc->scsi_done( sc );
            spin_unlock_irqrestore( &io_request_lock, cpu_flags );
            continue;
        }

        mycmd = AllocMyCmd( HostA, cie );
        if ( ! mycmd ) {
            DEBUG_ERR0("iSCSI: Out of local commands\n");
            NU_TRACE( NUSCSITCP_TRACE_CmdDone, sc->lun, sc->cmnd[0],
                sc->target, 0, 1 );
            sc->result = DID_BUS_BUSY << 16;
            sc->host_scribble = NULL;
            spin_lock_irqsave( &io_request_lock, cpu_flags );
            sc->scsi_done( sc );
            spin_unlock_irqrestore( &io_request_lock, cpu_flags );
            continue;
        }

        rc = NuScsiTcpXmitCmd( cie, sc, mycmd );
        if ( rc < 0 ) {
            FreeMyCmd( mycmd );

            if (rc == -EPIPE) {
                /* close the connection */
                cie->ConnFlag &= ~CONN_ALIVE;
                DEBUG_INIT1("iSCSI: connection 0x%x EPIPE\n", htonl(cie->RemoteIpAddr));
            }
            NU_TRACE( NUSCSITCP_TRACE_CmdDone, sc->lun, sc->cmnd[0],
                      sc->target, 0, 2 );
            sc->host_scribble = NULL;
            if ( sc->result == 0 ) {
                sc->result = DID_NO_CONNECT << 16;
                spin_lock_irqsave( &io_request_lock, cpu_flags );
                sc->scsi_done( sc );
                spin_unlock_irqrestore( &io_request_lock, cpu_flags );
            }
        } else {
            sc->host_scribble = (char *)mycmd;
            mycmd->sc = sc;
            mycmd->direction = rc;
            mycmd->cmdrn = cie->CmdRn++;
            if ( cie->CmdRn == 0 ) {
                cie->CmdRn = 1;
            }
            if ( (rc == NUSCSITCP_CMD_WRITE) && (cie->UseRtt == 0) ) {
                NuScsiTcpXmitImmData( cie, mycmd );
            }
        }
    }

    return;
}

static int
MyRecvMsg( ConnInfoEntry *cie, struct msghdr *msg, int len )
{
    int rc;
    mm_segment_t oldfs;

    oldfs = get_fs();
    set_fs( get_ds() );
    rc = sock_recvmsg( cie->TcpSock, msg, len, MSG_WAITALL );
    set_fs( oldfs );

    cie->RxBytes += len;

    return rc;
}

static int
MySendMsg( ConnInfoEntry *cie, struct msghdr *msg, int len )
{
    int rc;
    mm_segment_t oldfs;

    /* don't write to dying connections */
    if ((cie->ConnFlag & CONN_ALIVE) == 0) {
        printk("iSCSI: attempt to send to dying connection 0x%x\n", htonl(cie->RemoteIpAddr));
        return -1;
    }

    current->flags |= PF_MEMALLOC;
    oldfs = get_fs();
    set_fs( get_ds() );
    rc = sock_sendmsg( cie->TcpSock, msg, len );
    set_fs( oldfs );
    current->flags &= ~PF_MEMALLOC;

    cie->TxBytes += len;

    return rc;
}

int
NuScsiTcpConnTask( void *vtaskp )
{
    int rc;
    ConnInfoEntry *cie;
    struct socket *sock;
    Scsi_Cmnd *sc, *scq;
    NuScsiTcpCmdEntry *mycmd;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )
    unsigned long cpu_flags;
#else
    unsigned int cpu_flags;
#endif

    if ( ! vtaskp ) {
        DEBUG_ERR0("iSCSI: Task parameter NULL\n");
        return 0;
    }

    cie = (ConnInfoEntry *)vtaskp;
    sprintf(current->comm,"NuConnTask%d",cie->ConnI);
    sock = cie->TcpSock;

    DEBUG_INIT1("iSCSI: Connection-task, 0x%x\n",ntohl(cie->RemoteIpAddr));

    /* sleep til the Timer is running */
    while( ! HostA ) {
        if ( signal_pending(current) ) {
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        rc = schedule_timeout( 10 );
    }

    /* while timer running and connection is alive */
    while( HostA && TimerActive && (cie->ConnFlag & CONN_ALIVE) ) {
        if ( cie->ConnScHead == NULL ) {
            /* sleep when we have no commands for this connection */
            interruptible_sleep_on_timeout( &cie->ConnWaitQ, 10 );
        }

        /* have we been killed? */
        if ( signal_pending(current) ) {
            break;
        }

        /* R2T */
        if ( cie->UseRtt ) {
            mycmd = cie->CmdHead;
            while( mycmd ) {
                sc = mycmd->sc;
                if ( mycmd->senddata ) {
                    /*
                     * Check for Q'd commands to get them going as soon as possible.
                     */
                    down( &cie->ConnLock );
                    scq = cie->ConnScHead;
                    cie->ConnScHead = NULL;
                    up( &cie->ConnLock );
                    if ( scq ) {
                        NuScsiTcpXmitQdCmds( cie, scq );
                    }

                    NuScsiTcpXmitRttData( cie, mycmd );
                    mycmd->senddata = 0;
                }
                mycmd = mycmd->next;
            }
        }

        down( &cie->ConnLock );
        scq = cie->ConnScHead;
        cie->ConnScHead = NULL;
        up( &cie->ConnLock );
        if ( scq ) {
            NuScsiTcpXmitQdCmds( cie, scq );
        }

        if ((cie->ConnFlag & (CONN_ALIVE|CONN_PING)) == (CONN_ALIVE|CONN_PING)) {
            /* send a ping and expect a response */
            NuScsiTcpXmitPing( cie );
            cie->ConnFlag &= ~CONN_PING;
            cie->ConnFlag |= CONN_PINGRSP;
        }
    }

    DEBUG_INIT1("iSCSI: Connection-task connection down 0x%x\n",
        ntohl(cie->RemoteIpAddr));

    down( &cie->ConnLock );
    scq = cie->ConnScHead;
    cie->ConnScHead = NULL;
    up( &cie->ConnLock );
    while ( scq ) {
        sc = scq;
        scq = (Scsi_Cmnd *)sc->host_scribble;
        sc->host_scribble = NULL;
        if ( sc->result == 0 ) {
            printk("iSCSI: Abort command sc 0x%x\n",(unsigned int)sc);
            sc->result = DID_SOFT_ERROR << 16;
            spin_lock_irqsave( &io_request_lock, cpu_flags );
            sc->scsi_done( sc );
            spin_unlock_irqrestore( &io_request_lock, cpu_flags );
        }
    }

    if ( cie->ConnPid ) {
        DEBUG_INIT2("iSCSI: TaskPid %d killing ConnPid %d\n", cie->TaskPid, cie->ConnPid);
        kill_proc(cie->ConnPid, SIGPIPE, 1);
    }

    current->state = TASK_RUNNING;
    DEBUG_INIT1("iSCSI: Connection-task leaving kernel 0x%x\n",
                ntohl(cie->RemoteIpAddr));

    return 0;
}

void
NuScsiTcpXmitNop( ConnInfoEntry *cie, struct ScsiTcpNopInHdr *stnih )
{
    struct ScsiTcpNopOutHdr stnoh;
    struct msghdr msg;
    struct iovec iov;
    int rc;

    if ( (stnih == NULL) || stnih->poll ) {
        memset( &stnoh, 0, sizeof(stnoh) );
        stnoh.opcode = SCSI_TCP_OP_NOOP_OUT;
        stnoh.expstatrn = htonl(cie->ExpStatRn);

        iov.iov_base = &stnoh;
        iov.iov_len = sizeof(stnoh);
        memset( &msg, 0, sizeof(msg) );
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;

        rc = MySendMsg( cie, &msg, sizeof(stnoh) );
        if ( rc != sizeof(stnoh) ) {
            DEBUG_ERR1("iSCSI: XmitNop error %d\n",rc);
            /* close the connection */
            cie->ConnFlag &= ~CONN_ALIVE;
        }

        NU_TRACE( NUSCSITCP_TRACE_TxNop, 0, 0, 0, 0, cie->RemoteIpAddr );
    }

    return;
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4 
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
