/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: nuscsitcp.h,v 1.15.4.1 2001/05/14 19:23:54 mpm Exp $ 
 *
 * nuscsitcp.h
 *
 *    Main include for iSCSI driver
 * 
 * May 2000, MTM 
 *
 */

#define NUSCSITCP_MAX_TARGETS	8
#define NUSCSITCP_MAX_LUN		32

/*
 * ioctls
 */
#define NUSCSITCP_GO			0x00470301
#define NUSCSITCP_CONN		0x00470302
#define NUSCSITCP_GETTRACE	0x00470303

typedef struct NuScsiTcpIoctlData_t {
	int				TcpSock;
	int				NumTargets;
	int				UseRtt;
	unsigned int	RemoteIpAddr;
	int				TargetIds[NUSCSITCP_MAX_TARGETS];
	char				TargetNames[NUSCSITCP_MAX_TARGETS][16];
} NuScsiTcpIoctlData;

#define NUSCSITCP_TRACE_COUNT	100

typedef struct NuScsiTcpTraceEntry_t {
	unsigned char	flag;
	unsigned char	lun;
	unsigned char	cmd;
	unsigned char	target;
	unsigned long	itt;
	unsigned long	data;
	int				activecmds;
	unsigned long	jiffies;
} NuScsiTcpTraceEntry;

/*
 * Trace flags
 */
#define NUSCSITCP_TRACE_Qd						0x01
#define NUSCSITCP_TRACE_TxCmd					0x02
#define NUSCSITCP_TRACE_RxCmd					0x03
#define NUSCSITCP_TRACE_RxCmdStatus			0x13
#define NUSCSITCP_TRACE_RxCmdFlow			0x23
#define NUSCSITCP_TRACE_RxCmdCmdStatus		0x33
#define NUSCSITCP_TRACE_TxData				0x04
#define NUSCSITCP_TRACE_RxData				0x05
#define NUSCSITCP_TRACE_TxAbort				0x06
#define NUSCSITCP_TRACE_RxAbort				0x07
#define NUSCSITCP_TRACE_CmdDone				0x08
#define NUSCSITCP_TRACE_CmdAborted			0x09
#define NUSCSITCP_TRACE_RttData				0x0a
#define NUSCSITCP_TRACE_TxPing				0x0b
#define NUSCSITCP_TRACE_RxPing				0x0c
#define NUSCSITCP_TRACE_TxNop					0x0d

#ifdef NUSCSITCP_DAEMON


#else

#define NUSCSITCP_MAX_CMDS		12
#define NUSCSITCP_CANQUEUE		64
#define NUSCSITCP_MAX_SG		64
#define NUSCSITCP_RXCTRL_SIZE	1460
#define NUSCSITCP_TIMEOUT		(HZ*5)

#define NUSCSITCP_PROC_INFO_DEF nuscsitcp_proc_info

/*
 * Scsi_Host Template
 */
#define NUSCSITCP {                             \
    next : NULL,                                \
    module : NULL,                              \
    proc_dir : NULL,                            \
    proc_info : NUSCSITCP_PROC_INFO_DEF,        \
    name : NULL,                                \
    detect : nuscsitcp_detect,                  \
    release : nuscsitcp_release,                \
    info : nuscsitcp_info,                      \
    ioctl : NULL,                               \
    command : NULL,                             \
    queuecommand : nuscsitcp_queue,             \
    eh_strategy_handler : NULL,                 \
    eh_abort_handler : nuscsitcp_eh_abort,      \
    eh_device_reset_handler : nuscsitcp_eh_device_reset, \
    eh_bus_reset_handler : nuscsitcp_eh_bus_reset, \
    eh_host_reset_handler : nuscsitcp_eh_host_reset, \
    abort : NULL,                               \
    reset : NULL,                               \
    slave_attach : NULL,                        \
    bios_param : nuscsitcp_biosparam,           \
    can_queue : NUSCSITCP_CANQUEUE,             \
    this_id: -1,                                \
    sg_tablesize : NUSCSITCP_MAX_SG,				\
    cmd_per_lun: NUSCSITCP_MAX_CMDS,            \
    present : 0,                                \
    unchecked_isa_dma : 0,                      \
    use_clustering : ENABLE_CLUSTERING,         \
    use_new_eh_code : 1,                        \
    emulated : 1                                \
}

/*    sg_tablesize : SG_NONE,                     \*/
/*    sg_tablesize : SG_ALL,                      \*/

#define NUSCSITCP_CMD_CONTROL	0x00
#define NUSCSITCP_CMD_WRITE	0x01
#define NUSCSITCP_CMD_READ		0x02

typedef struct NuScsiTcpCmdEntry_t {
	struct NuScsiTcpCmdEntry_t	*next;
	struct NuScsiTcpCmdEntry_t	*prev;
	Scsi_Cmnd						*sc;
	struct ConnectionInfo		*cie;
	int								rxdata;
	unsigned int					rxrsp:1,
										rxovrflw:1,
										rxundflw:1,
										cmdaborted:1,
										abortdone:1,
										senddata:1,
										rsvd:26;
	int								direction;
	unsigned int					ttt;
	unsigned int					data_offset;
	unsigned int					cmdrn;
	unsigned long					jiffies;
} NuScsiTcpCmdEntry;

typedef struct NuScsiTcpHostAdapter_t {
	int						active;
	Scsi_Cmnd				*q_head;
	Scsi_Cmnd				*q_tail;
	NuScsiTcpCmdEntry		*cmd_mem;
	NuScsiTcpCmdEntry		*cmd_max;
	NuScsiTcpCmdEntry		*free_cmd_head;
	NuScsiTcpCmdEntry		*free_cmd_tail;
} NuScsiTcpHostAdapter;

typedef struct ConnectionInfo {
	struct ConnectionInfo	*Next;
	unsigned int				ConnFlag;
	Scsi_Cmnd					*ConnScHead;
	Scsi_Cmnd					*ConnScTail;
	NuScsiTcpCmdEntry			*CmdHead;
	NuScsiTcpCmdEntry			*CmdTail;
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) )  
	wait_queue_head_t			ConnWaitQ;
#else
	struct wait_queue			*ConnWaitQ;
#endif
	struct semaphore			ConnLock;
	unsigned int				RxBytes;
	unsigned int				TxBytes;
	unsigned int				CmdRn;
	unsigned int				ExpStatRn;
	unsigned int				PingCount;
	struct socket				*TcpSock;
	int							UseRtt;
	unsigned int				RemoteIpAddr;
	int							NumTargets;
	int							ConnPid;
	int							TaskPid;
	int							ConnI;
	unsigned long				Jiffies;
	unsigned long				PingTimeout;
	unsigned char				RxBuf[NUSCSITCP_RXCTRL_SIZE];
	struct iovec				RxIov[NUSCSITCP_MAX_SG];
	struct iovec				TxIov[(NUSCSITCP_MAX_SG+1)];
} ConnInfoEntry;

/*
 * Connection information flags
 */
#define CONN_ALIVE	0x01
#define CONN_PING		0x02
#define CONN_WAKE		0x04
#define CONN_PINGRSP	0x08
#define CONN_DEAD		0x10

#ifdef DEBUG_INIT
#define DEBUG_INIT0(DI0) printk(DI0);
#define DEBUG_INIT1(DI0,DI1) printk(DI0,DI1);
#define DEBUG_INIT2(DI0,DI1,DI2) printk(DI0,DI1,DI2);
#define DEBUG_INIT3(DI0,DI1,DI2,DI3) printk(DI0,DI1,DI2,DI3);
#define DEBUG_INIT4(DI0,DI1,DI2,DI3,DI4) printk(DI0,DI1,DI2,DI3,DI4);
#define DEBUG_INIT5(DI0,DI1,DI2,DI3,DI4,DI5) printk(DI0,DI1,DI2,DI3,DI4,DI5);
#else
#define DEBUG_INIT0(DI0)
#define DEBUG_INIT1(DI0,DI1)
#define DEBUG_INIT2(DI0,DI1,DI2)
#define DEBUG_INIT3(DI0,DI1,DI2,DI3)
#define DEBUG_INIT4(DI0,DI1,DI2,DI3,DI4)
#define DEBUG_INIT5(DI0,DI1,DI2,DI3,DI4,DI5)
#endif

#ifdef DEBUG_FLOW
#define DEBUG_FLOW0(DF0) printk(DF0);
#define DEBUG_FLOW1(DF0,DF1) printk(DF0,DF1);
#define DEBUG_FLOW2(DF0,DF1,DF2) printk(DF0,DF1,DF2);
#define DEBUG_FLOW3(DF0,DF1,DF2,DF3) printk(DF0,DF1,DF2,DF3);
#define DEBUG_FLOW4(DF0,DF1,DF2,DF3,DF4) printk(DF0,DF1,DF2,DF3,DF4);
#define DEBUG_FLOW5(DF0,DF1,DF2,DF3,DF4,DF5) printk(DF0,DF1,DF2,DF3,DF4,DF5);
#else
#define DEBUG_FLOW0(DF0)
#define DEBUG_FLOW1(DF0,DF1)
#define DEBUG_FLOW2(DF0,DF1,DF2)
#define DEBUG_FLOW3(DF0,DF1,DF2,DF3)
#define DEBUG_FLOW4(DF0,DF1,DF2,DF3,DF4)
#define DEBUG_FLOW5(DF0,DF1,DF2,DF3,DF4,DF5)
#endif

#ifdef DEBUG_ERROR
#define DEBUG_ERR0(DE0) printk(DE0);
#define DEBUG_ERR1(DE0,DE1) printk(DE0,DE1);
#define DEBUG_ERR2(DE0,DE1,DE2) printk(DE0,DE1,DE2);
#define DEBUG_ERR3(DE0,DE1,DE2,DE3) printk(DE0,DE1,DE2,DE3);
#define DEBUG_ERR4(DE0,DE1,DE2,DE3,DE4) printk(DE0,DE1,DE2,DE3,DE4);
#define DEBUG_ERR5(DE0,DE1,DE2,DE3,DE4,DE5) printk(DE0,DE1,DE2,DE3,DE4,DE5);
#else
#define DEBUG_ERR0(DE0)
#define DEBUG_ERR1(DE0,DE1)
#define DEBUG_ERR2(DE0,DE1,DE2)
#define DEBUG_ERR3(DE0,DE1,DE2,DE3)
#define DEBUG_ERR4(DE0,DE1,DE2,DE3,DE4)
#define DEBUG_ERR5(DE0,DE1,DE2,DE3,DE4,DE5)
#endif

#endif

/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 *
 * $Id: version.h,v 1.23.2.9 2001/07/10 20:31:44 sferris Exp $ 
 *
 *      controls the version number printed by the iSCSI driver
 *
 * May 2000, MTM
 *
 */

#define SCSI_TCP_DRIVER_VERSION	7
#define INTERNAL_DRIVER_VERSION	5
#define ISCSI_DATE	"10-Jul-2001"
#define BUILD_STR	NULL
