/*
 * Copyright (c) 2001 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 * Mountain View, CA  94043, or:
 *
 * http://www.sgi.com
 *
 * For further information regarding this notice, see:
 *
 * http://oss.sgi.com/projects/GenInfo/SGIGPLNoticeExplan/
 */
#ifndef _LINUX_ACL_H
#define _LINUX_ACL_H

typedef ushort	acl_perm_t;
typedef int	acl_type_t;
typedef int	acl_tag_t;

/*
 * Number of "base" ACL entries (USER_OBJ, GROUP_OBJ, MASK, & OTHER_OBJ)
 */
#define ACL_MAX_ENTRIES 25

/*
 * System call interface representation of an ACL.
 */
struct acl_entry {
	acl_tag_t 		ae_tag;
	uid_t			ae_id;
	acl_perm_t		ae_perm;	
};
struct acl {
	int			acl_cnt;	/* Number of entries */
	struct acl_entry	acl_entry[ACL_MAX_ENTRIES];
};

typedef struct acl_entry * acl_entry_t;
typedef struct acl * acl_t;

/* acl_tag_t */
#define ACL_UNDEFINED_TAG	0x00			/* undefined tag */
#define ACL_USER_OBJ		0x01			/* owner */
#define ACL_USER		0x02			/* additional users */
#define ACL_GROUP_OBJ		0x04			/* group */
#define ACL_GROUP		0x08			/* additional groups */
#define ACL_MASK		0x10			/* mask entry */
#define ACL_OTHER_OBJ		0x20			/* other entry */
#define ACL_OTHER		0x20			/* POSIX other entry */

/* acl_type_t */
#define ACL_TYPE_ACCESS         0x00
#define ACL_TYPE_DEFAULT        0x01

/* acl_perm_t */
#define ACL_PERM_NONE		0x00
#define ACL_READ		0x04
#define ACL_WRITE		0x02
#define ACL_EXECUTE		0x01

/*
 * An IRIX defined macro not in P1003.1e.
 * With ACL_COMPAT_IRIXGET set it used to signify an empty ACL.
 * It is also used to delete an ACL.
 */
#define ACL_NOT_PRESENT	-1

#endif /* __LINUX_ACL_H__ */
