#include <stdio.h>

#include <pnm.h>

#define MAX_COLORS	214
#define MAX_COLORS_16	16
#define MAX_COLORS_BW	2

int color_lookup(xel color,xel *colormap,int *used,int maxcolor)
{
	int i;

	for (i=0;i<*used;i++) {
		if (PPM_EQUAL(colormap[i],color))
			return i;
	}

	if (i >= maxcolor) {
		fprintf(stderr,"Too many colors: %d\n",maxcolor);
		exit(1);
	}

	(*used)++;
	colormap[i]=color;
	return i;
}

void print_data(int *data,int len,const char *prefix,const char *name)
{
	int i;

	printf("unsigned char linux_logo%s%s[] __initdata = {",prefix,name);

	for (i=0;i<len;i++) {
		if ( (i%8) == 0) printf("\n ");
		printf(" 0x%.2X,",data[i]);
	}

	printf("\n};\n\n");
}

void linux_encode_bw(xel **ppm,int rows,int cols,int colormax,const char *prefix)
{
	int *m;
	int row,col,i,used=0,bytes;
	xel black;

	PPM_PUTR(black,0);
	PPM_PUTG(black,0);
	PPM_PUTB(black,0);

	bytes=(cols+7)>>3;
	m=malloc(sizeof(int)*rows*bytes);
	for (row=i=0;row<rows;row++) {
		for (col=0;col<cols;col++) {
			if ((col&7)==0) {
				m[++i] = 0;
			} else {
				m[i] <<= 1;
			}
			if (!PPM_EQUAL(ppm[row][col],black))
				m[i] |= 1;
		}
		/* End of row... */
		for (;(col&7) != 0;col++) {
			m[i] <<= 1;
		}
	}

	printf("#ifdef INCLUDE_LINUX_LOGOBW\n\n");

	print_data(m,rows*bytes,prefix,"");

	printf("\n#endif /* INCLUDE_LINUX_LOGOBW */\n\n");

	free(m);
}

void linux_encode_16(xel **ppm,int rows,int cols,int colormax,const char *prefix)
{
	int *r,*g,*b,*m;
	xel colormap[MAX_COLORS];
	int row,col,i,used=0;

	m=malloc(sizeof(int)*rows*((cols+1)>>1));
	for (row=i=0;row<rows;row++) {
		for (col=0;col<cols;col++) {
			if ((col&1)==0) {
				m[++i]=0;
			} else {
				m[i] <<= 4;
			}
			m[i] |= color_lookup(ppm[row][col],colormap,&used,colormax);
		}
	}

	r=malloc(sizeof(int)*used);
	g=malloc(sizeof(int)*used);
	b=malloc(sizeof(int)*used);

	for (i=0;i<used;i++) {
		r[i]=PPM_GETR(colormap[i]);
		g[i]=PPM_GETG(colormap[i]);
		b[i]=PPM_GETB(colormap[i]);
	}

	printf("#ifdef INCLUDE_LINUX_LOGO%s\n\n",prefix);

	print_data(r,used,prefix,"_red");
	print_data(g,used,prefix,"_green");
	print_data(b,used,prefix,"_blue");
	print_data(m,rows*((cols+1)>>1),prefix,"");

	printf("\n#endif /* INCLUDE_LINUX_LOGO%s */\n\n",prefix);

	free(r);
	free(g);
	free(b);
	free(m);
}

void linux_encode(xel **ppm,int rows,int cols,int colormax,const char *prefix)
{
	int *r,*g,*b,*m;
	xel colormap[MAX_COLORS];
	int row,col,i,used=0;

	m=malloc(sizeof(int)*rows*cols);
	for (row=i=0;row<rows;row++) {
		for (col=0;col<cols;col++,i++) {
			m[i]=color_lookup(ppm[row][col],colormap,&used,colormax)+32;
		}
	}

	r=malloc(sizeof(int)*used);
	g=malloc(sizeof(int)*used);
	b=malloc(sizeof(int)*used);

	for (i=0;i<used;i++) {
		r[i]=PPM_GETR(colormap[i]);
		g[i]=PPM_GETG(colormap[i]);
		b[i]=PPM_GETB(colormap[i]);
	}

	printf("#define LOGO_W %d\n",cols);
	printf("#define LOGO_H %d\n\n",rows);

	print_data(r,used,prefix,"_red");
	print_data(g,used,prefix,"_green");
	print_data(b,used,prefix,"_blue");
	print_data(m,rows*cols,prefix,"");

	free(r);
	free(g);
	free(b);
	free(m);
}

int main(int argc,char **argv)
{
	xelval maxval;
	xel **ppm;
	FILE *inf;
	int rows,cols,format,colors,colormax;
	xel colormap[MAX_COLORS];
	const char *prefix;
	void (*encode)(xel **ppm,int rows,int cols,int colormax,const char *prefix);

	if (argc>1 && strcmp(argv[1],"-16")==0) {
		argc--;
		memcpy(argv+1,argv+2,(argc-1)*sizeof(argv[0]));
		prefix="16";
		colormax=MAX_COLORS_16;
		encode=linux_encode_16;
	} else if (argc>1 && strcmp(argv[1],"-bw")==0) {
		argc--;
		memcpy(argv+1,argv+2,(argc-1)*sizeof(argv[0]));
		prefix="_bw";
		colormax=MAX_COLORS_BW;
		encode=linux_encode_bw;
	} else {
		prefix="";
		colormax=MAX_COLORS;
		encode=linux_encode;
	}

	if (argc==1) {
		inf=stdin;
	} else {
		inf=fopen(argv[1],"r");
		if (inf==NULL)
			exit(1);
	}

	ppm=pnm_readpnm(inf,&cols,&rows,&maxval,&format);
	if (format != RPPM_FORMAT &&
	    format != PPM_FORMAT) {
		fprintf(stderr,"%s: Not a ppm file!\n",argv[1]);
		exit(1);
	}

	encode(ppm,rows,cols,colormax,prefix);

	return 0;
}
